@@ -46,11 +46,28 @@ def mkdirSafe(directory):
             print('Directory "' + directory + '" already exists!')
 
 
-def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False, disableMeshOptimization=False):
+def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False,
+                 disableMeshOptimization=False, enableMultiFile=False, staticBase=False, toolSlot=None,
+                 initRotation='0 1 0 0', initPos=None, linkToDef=False, jointToDef=False):
     if not os.path.exists(inFile):
         sys.exit('Input file "%s" does not exists.' % inFile)
+    if not type(initRotation) == str or len(initRotation.split()) != 4:
+        sys.exit('--rotation argument is not valid. Has to be of Type = str and contain 4 values.')
+    if initPos is not None:
+        try:
+            initPos = initPos.replace(",", ' ').replace("[", '').replace("]", '').replace("(", '').replace(")", '')
+            initPos = list(map(float, initPos.split()))
+        except Exception as e:
+            sys.exit(e, '\n--init-pos argument is not valid. Your list has to be inside of quotation marks. '
+                     'Example: --init-pos="1.0, 2, -0.4"')
 
     urdf2webots.parserURDF.disableMeshOptimization = disableMeshOptimization
+    urdf2webots.writeProto.enableMultiFile = enableMultiFile
+    urdf2webots.writeProto.staticBase = staticBase
+    urdf2webots.writeProto.toolSlot = toolSlot
+    urdf2webots.writeProto.initPos = initPos
+    urdf2webots.writeProto.linkToDef = linkToDef
+    urdf2webots.writeProto.jointToDef = jointToDef
 
     with open(inFile, 'r') as file:
         inPath = os.path.dirname(os.path.abspath(inFile))
@@ -58,7 +75,7 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False, disable
         packages = re.findall('"package://(.*)"', content)
         if packages:
             packageName = packages[0].split('/')[0]
-            directory = os.path.dirname(inFile)
+            directory = inPath
             while packageName != os.path.split(directory)[1] and os.path.split(directory)[1]:
                 directory = os.path.dirname(directory)
             if not os.path.split(directory)[1]:
@@ -80,12 +97,26 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False, disable
 
         for child in domFile.childNodes:
             if child.localName == 'robot':
-                robotName = convertLUtoUN(urdf2webots.parserURDF.getRobotName(child))  # capitalize
-                outputFile = outFile if outFile else robotName + '.proto'
+                if outFile:
+                    if os.path.splitext(os.path.basename(outFile))[1] == '.proto':
+                        robotName = os.path.splitext(os.path.basename(outFile))[0]
+                        outputFile = outFile
+                    else:
+                        # treat outFile as directory and construct filename
+                        robotName = convertLUtoUN(urdf2webots.parserURDF.getRobotName(child))  # capitalize
+                        outputFile = os.path.join(outFile, robotName + '.proto')
+                else:
+                    robotName = convertLUtoUN(urdf2webots.parserURDF.getRobotName(child))  # capitalize
+                    outputFile = outFile if outFile else robotName + '.proto'
+                urdf2webots.writeProto.robotNameMain = robotName
 
                 urdf2webots.parserURDF.robotName = robotName  # pass robotName
                 mkdirSafe(outputFile.replace('.proto', '') + '_textures')  # make a dir called 'x_textures'
 
+                if enableMultiFile:
+                    mkdirSafe(outputFile.replace('.proto', '') + '_meshes')  # make a dir called 'x_meshes'
+                    urdf2webots.writeProto.meshFilesPath = outputFile.replace('.proto', '') + '_meshes'
+
                 robot = child
                 protoFile = open(outputFile, 'w')
                 urdf2webots.writeProto.header(protoFile, inFile, robotName)
@@ -149,7 +180,7 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False, disable
                               urdf2webots.parserURDF.Lidar.list)
                 print('There are %d links, %d joints and %d sensors' % (len(linkList), len(jointList), len(sensorList)))
 
-                urdf2webots.writeProto.declaration(protoFile, robotName)
+                urdf2webots.writeProto.declaration(protoFile, robotName, initRotation)
                 urdf2webots.writeProto.URDFLink(protoFile, rootLink, 1, parentList, childList, linkList, jointList,
                                                 sensorList, boxCollision=boxCollision, normal=normal, robot=True)
                 protoFile.write('}\n')
@@ -161,7 +192,8 @@ def convert2urdf(inFile, outFile=None, normal=False, boxCollision=False, disable
 if __name__ == '__main__':
     optParser = optparse.OptionParser(usage='usage: %prog --input=my_robot.urdf [options]')
     optParser.add_option('--input', dest='inFile', default='', help='Specifies the urdf file to convert.')
-    optParser.add_option('--output', dest='outFile', default='', help='Specifies the name of the resulting PROTO file.')
+    optParser.add_option('--output', dest='outFile', default='', help='Specifies the path and, if ending in ".proto", name '
+                         'of the resulting PROTO file. The filename minus the .proto extension will be the robot name.')
     optParser.add_option('--normal', dest='normal', action='store_true', default=False,
                          help='If set, the normals are exported if present in the URDF definition.')
     optParser.add_option('--box-collision', dest='boxCollision', action='store_true', default=False,
@@ -169,6 +201,22 @@ if __name__ == '__main__':
     optParser.add_option('--disable-mesh-optimization', dest='disableMeshOptimization', action='store_true', default=False,
                          help='If set, the duplicated vertices are not removed from the meshes (this can speed up a lot the '
                          'conversion).')
+    optParser.add_option('--multi-file', dest='enableMultiFile', action='store_true', default=False,
+                         help='If set, the mesh files are exported as separated PROTO files')
+    optParser.add_option('--static-base', dest='staticBase', action='store_true', default=False,
+                         help='If set, the base link will have the option to be static (disable physics)')
+    optParser.add_option('--tool-slot', dest='toolSlot', default=None,
+                         help='Specify the link that you want to add a tool slot too (exact link name from urdf)')
+    optParser.add_option('--rotation', dest='initRotation', default='0 1 0 0',
+                         help='Set the rotation field of your PROTO file.')
+    optParser.add_option('--init-pos', dest='initPos', default=None,
+                         help='Set the initial positions of your robot joints. Example: --init-pos="[1.2, 0.5, -1.5]" would '
+                         'set the first 3 joints of your robot to the specified values, and leave the rest with their '
+                         'default value')
+    optParser.add_option('--link-to-def', dest='linkToDef', action='store_true', default=False,
+                         help='If set, urdf link names are also used as DEF names as well as solid names.')
+    optParser.add_option('--joint-to-def', dest='jointToDef', action='store_true', default=False,
+                         help='If set, urdf joint names are also used as DEF names as well as joint names.')
     options, args = optParser.parse_args()
-
-    convert2urdf(options.inFile, options.outFile, options.normal, options.boxCollision, options.disableMeshOptimization)
+    convert2urdf(options.inFile, options.outFile, options.normal, options.boxCollision, options.disableMeshOptimization,
+                 options.enableMultiFile, options.staticBase, options.toolSlot, options.initRotation, options.initPos, options.linkToDef, options.jointToDef)
